# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>

from libtuning.modules.agc.agc import AGC
from libtuning.modules.agc.rkisp1 import AGCRkISP1
